/* maxflow.cpp */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bk.h"

#ifdef GRAPH_NAMESPACE
namespace GRAPH_NAMESPACE {
#endif

#define INFINITE_D ((int)(((unsigned)-1)/2))		/* infinite distance to the terminal */

/***********************************************************************/

/*
	Functions for processing active list.
	i->next points to the next node in the list
	(or to i, if i is the last node in the list).
	If i->next is NULL iff i is not in the list.

	There are two queues. Active nodes are added
	to the end of the second queue and read from
	the front of the first queue. If the first queue
	is empty, it is replaced by the second queue
	(and the second queue becomes empty).
*/

/***********************************************************************/

inline void Graph::add_to_changed_list(node *i)
{
	if (changed_list && !i->is_in_changed_list)
	{
		node_id* ptr = changed_list->New();
		*ptr = (node_id)(i - nodes);
		i->is_in_changed_list = true;
	}
}

/***********************************************************************/

void Graph::maxflow_init()
{
	int _i;

	L.reset();

	TIME = 0;

	FOR_ALL_NODES(_i)
	{
		node* i = nodes + _i;
		i -> flag = FLAG_NONE;
		i -> is_in_changed_list = 0;
		i -> TS = TIME;
		if (i->excess > 0)
		{
			/* i is connected to the source */
			i -> is_sink = 0;
			i -> parent = TERMINAL;
			L.add_to_active_rear(i);
			i -> DIST = 1;
		}
		else if (i->excess < 0)
		{
			/* i is connected to the sink */
			i -> is_sink = 1;
			i -> parent = TERMINAL;
			L.add_to_active_rear(i);
			i -> DIST = 1;
		}
		else
		{
			i -> parent = NULL;
		}
	}
}

void Graph::maxflow_reuse_trees_init()
{
	node* i;
	node* j;
	arc* a;

	TIME ++;
	while ((i=L.next_marked()))
	{
#ifdef GRAPH_CONTRACTION
		assert (!i->is_contracted);
#endif

		if (i->excess == 0)
		{
			if (i->parent) L.add_to_orphan_rear(i);
			continue;
		}

		if (i->excess > 0)
		{
			if (!i->parent || i->is_sink)
			{
				i->is_sink = 0;
				L.add_to_active_rear(i);
				FOR_ALL_ARCS(i, a)
				{
					j = a->head;
					if (j->flag == FLAG_MARKED) continue;
					if (j->parent == REV_ARC(a)) L.add_to_orphan_rear(j);
					else if (j->flag == FLAG_NONE && j->parent && j->is_sink && IS_NONSATURATED(a)) L.add_to_active_rear(j);
				}
				add_to_changed_list(i);
			}
		}
		else
		{
			if (!i->parent || !i->is_sink)
			{
				i->is_sink = 1;
				L.add_to_active_rear(i);
				FOR_ALL_ARCS(i, a)
				{
					j = a->head;
					if (j->flag == FLAG_MARKED) continue;
					if (j->parent == REV_ARC(a)) L.add_to_orphan_rear(j);
					if (j->flag == FLAG_NONE && j->parent && !j->is_sink && IS_REV_NONSATURATED(a)) L.add_to_active_rear(j);
				}
				add_to_changed_list(i);
			}
		}
		i->parent = TERMINAL;
		i -> TS = TIME;
		i -> DIST = 1;
#ifdef ONE_SIDED_ACTIVE
		if (!i->flag)
		{
			L.add_to_active_rear(i);
		}
#endif
	}

	//test_consistency();

	/* adoption */
	while ((i=L.next_orphan()))
	{
		if (i->is_sink) process_sink_orphan(i);
		else            process_source_orphan(i);
#ifdef ONE_SIDED_ACTIVE
		if (i->parent && i->flag != FLAG_ACTIVE) L.add_to_active_rear(i);
#endif
	}
	/* adoption end */
	//test_consistency();
}

void Graph::augment(arc *middle_arc)
{
	node *i;
	arc *a;
	tcaptype bottleneck;

	/* 1. Finding bottleneck capacity */
	/* 1a - the source tree */
	bottleneck = GET_RCAP(middle_arc);
	for (i=REV_ARC(middle_arc)->head; ; i=a->head)
	{
		a = i -> parent;
		if (a == TERMINAL) break;
		if (bottleneck > GET_RCAP(REV_ARC(a))) bottleneck = GET_RCAP(REV_ARC(a));
	}
	if (bottleneck > i->excess) bottleneck = i->excess;
	/* 1b - the sink tree */
	for (i=middle_arc->head; ; i=a->head)
	{
		a = i -> parent;
		if (a == TERMINAL) break;
		if (bottleneck > GET_RCAP(a)) bottleneck = GET_RCAP(a);
	}
	if (bottleneck > -i->excess) bottleneck = -i->excess;


	/* 2. Augmenting */
	/* 2a - the source tree */
	PUSH_FLOW(middle_arc, bottleneck);
	for (i=REV_ARC(middle_arc)->head; ; i=a->head)
	{
		a = i -> parent;
		if (a == TERMINAL) break;
		PUSH_FLOW(a, -bottleneck);
		if (!IS_REV_NONSATURATED(a))
		{
			L.add_to_orphan_front(i); // add i to the beginning of the adoption list
		}
	}
	SET_EXCESS(i, i->excess - bottleneck);
	if (i->excess == 0)
	{
		L.add_to_orphan_front(i); // add i to the beginning of the adoption list
	}
	/* 2b - the sink tree */
	for (i=middle_arc->head; ; i=a->head)
	{
		a = i -> parent;
		if (a == TERMINAL) break;
		PUSH_FLOW(a, bottleneck);
		if (!IS_NONSATURATED(a))
		{
			L.add_to_orphan_front(i); // add i to the beginning of the adoption list
		}
	}
	SET_EXCESS(i, i->excess + bottleneck);
	if (i->excess == 0)
	{
		L.add_to_orphan_front(i); // add i to the beginning of the adoption list
	}


	flow += bottleneck;
}

/***********************************************************************/

void Graph::process_source_orphan(node *i)
{
	node *j;
	arc *a0, *a0_min = NULL, *a;
	int d, d_min = INFINITE_D;

	/* trying to find a new parent */
	FOR_ALL_ARCS(i, a0)
	if (IS_REV_NONSATURATED(a0))
	{
		j = a0 -> head;
		if (!j->is_sink && (a=j->parent))
		{
			/* checking the origin of j */
			d = 0;
			while ( 1 )
			{
				if (j->TS == TIME)
				{
					d += j -> DIST;
					break;
				}
				a = j -> parent;
				d ++;
				if (a==TERMINAL)
				{
					j -> TS = TIME;
					j -> DIST = 1;
					break;
				}
				if (a==ORPHAN) { d = INFINITE_D; break; }
				j = a -> head;
			}
			if (d<INFINITE_D) /* j originates from the source - done */
			{
				if (d<d_min)
				{
					a0_min = a0;
					d_min = d;
				}
				/* set marks along the path */
				for (j=a0->head; j->TS!=TIME; j=j->parent->head)
				{
					j -> TS = TIME;
					j -> DIST = d --;
				}
			}
		}
	}

	if ((i->parent = a0_min))
	{
		i -> TS = TIME;
		i -> DIST = d_min + 1;
		L.add_to_active_rear(i);
	}
	else
	{
		/* no parent is found */
		add_to_changed_list(i);

		/* process neighbors */
		FOR_ALL_ARCS(i, a0)
		{
			j = a0 -> head;
			if (!j->is_sink && (a=j->parent))
			{
				if (IS_REV_NONSATURATED(a0) && j->flag == FLAG_NONE)
				{
					L.add_to_active_rear(j);
				}
				if (a!=TERMINAL && a!=ORPHAN && a->head==i)
				{
					L.add_to_orphan_rear(j); // add j to the end of the adoption list
				}
			}
		}
	}
}

void Graph::process_sink_orphan(node *i)
{
	node *j;
	arc *a0, *a0_min = NULL, *a;
	int d, d_min = INFINITE_D;

	/* trying to find a new parent */
	FOR_ALL_ARCS(i, a0)
	if (IS_NONSATURATED(a0))
	{
		operation_count ++;
		j = a0 -> head;
		if (j->is_sink && (a=j->parent))
		{
			/* checking the origin of j */
			d = 0;
			while ( 1 )
			{
				if (j->TS == TIME)
				{
					d += j -> DIST;
					break;
				}
				a = j -> parent;
				d ++;
				if (a==TERMINAL)
				{
					j -> TS = TIME;
					j -> DIST = 1;
					break;
				}
				if (a==ORPHAN) { d = INFINITE_D; break; }
				j = a -> head;
			}
			if (d<INFINITE_D) /* j originates from the sink - done */
			{
				if (d<d_min)
				{
					a0_min = a0;
					d_min = d;
				}
				/* set marks along the path */
				for (j=a0->head; j->TS!=TIME; j=j->parent->head)
				{
					j -> TS = TIME;
					j -> DIST = d --;
				}
			}
		}
	}

	if ((i->parent = a0_min))
	{
		i -> TS = TIME;
		i -> DIST = d_min + 1;
		L.add_to_active_rear(i);
	}
	else
	{
		/* no parent is found */
		add_to_changed_list(i);

		/* process neighbors */
		FOR_ALL_ARCS(i, a0)
		{
			j = a0 -> head;
			if (j->is_sink && (a=j->parent))
			{
				if (IS_NONSATURATED(a0) && j->flag == FLAG_NONE)
				{
					L.add_to_active_rear(j);
				}
				if (a!=TERMINAL && a!=ORPHAN && a->head==i)
				{
					L.add_to_orphan_rear(j); // add j to the end of the adoption list
				}
			}
		}
	}
}

/***********************************************************************/

Graph::flowtype Graph::maxflow(bool reuse_trees, Block<node_id>* _changed_list)
{
	init_graph();

	node *i, *j, *current_node = NULL;
	arc *a;

	changed_list = _changed_list;
	if (maxflow_iteration == 0 && reuse_trees) { printf("reuse_trees cannot be used in the first call to maxflow()!"); exit(1); }
	if (changed_list && !reuse_trees) { printf("changed_list cannot be used without reuse_trees!"); exit(1); }

	if (reuse_trees) maxflow_reuse_trees_init();
	else             maxflow_init();

	int operation_count_prev = operation_count;

	// main loop
	while ( 1 ) 
	{
		i = L.next_active(); // this removes i from the queue
		if (!i) break;
		node* a_tail = NULL;

		// test_consistency(current_node);
		/* growth */
		if (!i->is_sink)
		{
			/* grow source tree */
			FOR_ALL_ARCS(i, a)
			if (IS_NONSATURATED(a))
			{
				operation_count ++;
				j = a -> head;
				if (!j->parent)
				{
					j -> is_sink = 0;
					j -> parent = REV_ARC(a);
					j -> TS = i -> TS;
					j -> DIST = i -> DIST + 1;
					L.add_to_active_rear(j);
					add_to_changed_list(j);
				}
				else if (j->is_sink) { a_tail = i; break; }
				else if (j->TS <= i->TS &&
				         j->DIST > i->DIST)
				{
					/* heuristic - trying to make the distance from j to the source shorter */
					j -> parent = REV_ARC(a);
					j -> TS = i -> TS;
					j -> DIST = i -> DIST + 1;
				}
			}
		}
		else
		{
			/* grow sink tree */
			FOR_ALL_ARCS(i, a)
			if (IS_REV_NONSATURATED(a))
			{
				operation_count ++;
				j = a -> head;
				if (!j->parent)
				{
					j -> is_sink = 1;
					j -> parent = REV_ARC(a);
					j -> TS = i -> TS;
					j -> DIST = i -> DIST + 1;
					L.add_to_active_rear(j);
					add_to_changed_list(j);
				}
				else if (!j->is_sink) { a_tail = j; a = REV_ARC(a); break; }
				else if (j->TS <= i->TS &&
				         j->DIST > i->DIST)
				{
					/* heuristic - trying to make the distance from j to the sink shorter */
					j -> parent = REV_ARC(a);
					j -> TS = i -> TS;
					j -> DIST = i -> DIST + 1;
				}
			}
		}

		TIME ++;

		if (a_tail)
		{
			L.add_to_active_front(i);

			/* augmentation */
			augment(a);
			/* augmentation end */

			if (!changed_list && operation_count > operation_count_prev + 2*edge_num && 0)
			{
				maxflow_init();
				operation_count_prev = operation_count;
			}
			else
			{
				/* adoption */
				while ((i=L.next_orphan()))
				{
					Lists::OrphanListData D;
					L.save_orphan_list(D);
					L.add_to_orphan_rear(i);
					while ((i=L.next_orphan()))
					{
						if (i->is_sink) process_sink_orphan(i);
						else            process_source_orphan(i);
					}
					L.restore_orphan_list(D);
				}
				/* adoption end */
			}
		}
	}
	// test_consistency();

	maxflow_iteration ++;

//#define CHECK_OPTIMALITY
#ifdef  CHECK_OPTIMALITY
	int _i;
	FOR_ALL_NODES(_i)
	{
		i = nodes + _i;
		int s = what_segment((int)(i-nodes));
		if (s == 0 && i->excess < 0) { printf("non-opimal flow! (#1)\n"); exit(1); }
		if (s == 1 && i->excess > 0) { printf("non-opimal flow! (#2)\n"); exit(1); }
		if (s == 1) continue;
		FOR_ALL_ARCS(i, a)
		{
			if (IS_NONSATURATED(a) && what_segment((int)(a->head-nodes)) != 0)
			{
				printf("non-opimal flow! (#3)\n"); exit(1);
			}
		}
	}
#endif

	return flow;
}

/***********************************************************************/


void Graph::test_consistency(node* current_node)
{
	node *i;
	arc *a;

#ifdef AFA
	int r;
	int num1 = 0, num2 = 0;

	// test whether all nodes i with i->next!=NULL are indeed in the queue
	for (i=nodes; i<node_last; i++)
	{
		if (i->next || i==current_node) num1 ++;
	}
	for (r=0; r<3; r++)
	{
		i = (r == 2) ? current_node : queue_first[r];
		if (i)
		for ( ; ; i=i->next)
		{
			num2 ++;
			if (i->next == i)
			{
				if (r<2) assert(i == queue_last[r]);
				else     assert(i == current_node);
				break;
			}
		}
	}
	assert(num1 == num2);
#endif

	int _i;
	FOR_ALL_NODES(_i)
	{
		i = nodes + _i;
		// test whether all edges in seach trees are non-saturated
		if (i->parent == NULL) {}
		else if (i->parent == ORPHAN) {}
		else if (i->parent == TERMINAL)
		{
			if (!i->is_sink) assert(i->excess > 0);
			else             assert(i->excess < 0);
		}
		else
		{
			if (!i->is_sink) assert (REV_ARC(i->parent)->r_cap > 0);
			else             assert (i->parent->r_cap > 0);
		}
		// test whether passive nodes in search trees have neighbors in
		// a different tree through non-saturated edges
		if (i->parent)
		{
			if (!i->is_sink && i->flag != FLAG_ACTIVE)
			{
				assert(i->excess >= 0);
				FOR_ALL_ARCS(i, a)
				{
					if (IS_NONSATURATED(a)) assert(a->head->parent && !a->head->is_sink);
				}
			}
			if (i->is_sink && !i->flag != FLAG_ACTIVE)
			{
				assert(i->excess <= 0);
				FOR_ALL_ARCS(i, a)
				{
					if (IS_REV_NONSATURATED(a)) assert(a->head->parent && a->head->is_sink);
				}
			}
		}
		// test marking invariants
		if (i->parent && i->parent!=ORPHAN && i->parent!=TERMINAL)
		{
			assert(i->TS <= i->parent->head->TS);
			if (i->TS == i->parent->head->TS) assert(i->DIST > i->parent->head->DIST);
		}
	}
}

#ifdef GRAPH_NAMESPACE
}
#endif

