#define _CRT_SECURE_NO_WARNINGS

#define TMP_GRAPH_IO

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gomory_hu.h"
#include "tmp_graph.h"
#include "run_maxflow.h"



void Usage(const char* s)
{
	printf("Usage: %s [options] <filename>\n", s);
	printf("options:\n");
	printf(" -s<0>      terminal selection method:\n");
	printf("            0: RANDOM-RANDOM (both terminals at random)\n");
	printf("            1: RANDOM-CLOSEST (first terminal at random, second closest to the first terminal)\n");
	printf("            2: HEAVIEST-HEAVEST (two terminals with largest weighted degree)\n");
	printf("            3: HEAVIEST-CLOSEST\n");
	printf("            4: CLOSEST-PAIR (similar to Akibo et al. 2016)\n");
	printf("            5: HEAVIEST-HEAVEST-EFFICIENT (<freq> is then ignored)\n");
	printf(" -f<freq>   frequency of updating s-t pairs (0 means infinity)\n");
	printf(" -o         save result to <filename>.tree in DIMACS format\n");
	printf(" -O         save result to <filename>.tree in plain format\n");

	printf("\n experimental:\n");
	printf(" -T:t0:t1:...:tk   use multistage algorithm with nondecreasing thresholds t0,t1,...,tk\n");
	printf(" -A:a0:a1:...:ak   similar to -T, but use ti=a0*(1+max weighted node degree)\n");
	printf(" -D:c0:f0:...:ck:fk   use multistage algorithm with adaptive thresholds: if the fraction of special nodes component in Gomory-Hu");
	printf("                      is less than ck then multiply current threshold by factor fk\n");

	exit(1);
}

// read string of the form a:b:c
int read_seq(const char* s, int* arr, int arr_size_max)
{
	int n=0;
	const char* s0 = s;
	while (*s && n<arr_size_max)
	{
		int i;
		for (i=0; s[i] && s[i] != ':'; i++) {}
		if (s[i] == ':')
		{
			if (sscanf(s, "%d:", &arr[n ++]) != 1) { printf("cannot parse string %s\n", s0); exit(1); }
			s += i+1;
		}
		else
		{
			if (sscanf(s, "%d", &arr[n ++]) != 1) { printf("cannot parse string %s\n", s0); exit(1); }
			break;
		}
	}
	return n;
}

int read_seq(const char* s, double* arr, int arr_size_max)
{
	int n=0;
	const char* s0 = s;
	while (*s && n<arr_size_max)
	{
		int i;
		for (i=0; s[i] && s[i] != ':'; i++) {}
		if (s[i] == ':')
		{
			if (sscanf(s, "%lf:", &arr[n ++]) != 1) { printf("cannot parse string %s\n", s0); exit(1); }
			s += i+1;
		}
		else
		{
			if (sscanf(s, "%lf", &arr[n ++]) != 1) { printf("cannot parse string %s\n", s0); exit(1); }
			break;
		}
	}
	return n;
}

int main(int argc, char* argv[])
{
	int i;
	TmpGraph<int>* g = NULL;

	GomoryHuOptions options;
	const int STAGE_MAX = 100;
	double D_array[STAGE_MAX];
	GH_REAL T_array[STAGE_MAX];
	double A_array[STAGE_MAX];
	const char* filename = NULL;
	int saveresult = 0;

	for (i=1; i<argc; i++)
	{
		const char* s = argv[i];
		if (s[0] == '-')
		{
			if      (s[1] == 'D' && s[2]==':') { options.method = GomoryHuOptions::METHOD_D; options.stage_num = read_seq(s+3, D_array, STAGE_MAX)/2; options.D_array = D_array; }
			else if (s[1] == 'T' && s[2]==':') { options.method = GomoryHuOptions::METHOD_T; options.stage_num = read_seq(s+3, T_array, STAGE_MAX); options.T_array = T_array; }
			else if (s[1] == 'A' && s[2]==':') { options.method = GomoryHuOptions::METHOD_A; options.stage_num = read_seq(s+3, A_array, STAGE_MAX); options.A_array = A_array; }
			else if (s[1] == 's') options.st_selection = (GomoryHuOptions::stSelection) atoi(s+2);
			else if (s[1] == 'f') options.st_freq = atof(s+2);
			else if (s[1] == 'v') options.verbose = atoi(s+2);
			else if (s[1] == 'o') saveresult = 1;
			else if (s[1] == 'O') saveresult = 2;
			else Usage(argv[0]);
		}
		else
		{
			if (filename) Usage(argv[0]);
			filename = s;
		}
	}

	if (!filename) Usage(argv[0]);

	int source = -1, sink = -1;
	g = ReadDIMACS(filename, &source, &sink);

	if (source >= 0 && sink >= 0)
	{
		RunMaxflow(g, source, sink);
		delete g;
		return 0;
	}
	
	int* parents = new int[2*g->n];
	int* cut_costs = parents + g->n;

	SolveGomoryHu(g, parents, cut_costs, options);
	if (saveresult)
	{
		char SAVEFILE[100];
		snprintf(SAVEFILE, sizeof(SAVEFILE), "%s.tree", filename);
		TmpGraph<int> solution(g->n, g->n-1);
		for (i=0; i<g->n; i++)
		{
			if (parents[i] >= 0) solution.AddEdge(i, parents[i], cut_costs[i]);
		}	
		if (saveresult == 1) SaveDIMACS(&solution, SAVEFILE);
		else                 SavePlain(&solution, SAVEFILE);
	}

	delete [] parents;
	delete g;

	return 0;
}

