#ifndef __GRAPH_WITH_TREE__
#define __GRAPH_WITH_TREE__

#include "gomory_hu.h"
#include "block.h"

class GomoryHuNested;
class GraphWithTreeAndLCA;

class GraphWithTree
{
public:
	GraphWithTree(int node_num);
	~GraphWithTree();

	void AddTreeEdge(int i, int p);
	void AddRoot(int r); // must be called last

	void AddEdge(int _i, int _j, GH_REAL w);

	void StartNode(int i) { i_current = i; }
	bool AddEdge(int j, GH_REAL w);
	void FinishNode();
	void MergeParallelEdges();

	void Print()
	{
		Node* i;
		Arc* a;
		for (i=nodes; i<nodes+node_num; i++)
		for (a=i->first; a; a=a->next)
		if (a->head > i)
		{
			printf("%d-%d[%d] ", (int)(i-nodes), (int)(a->head-nodes), (int)a->w);
		}
		printf("\n");
	}

protected:
	friend class GomoryHuNested;
	friend class GraphWithTreeAndLCA;

	struct Arc;
	struct Node
	{
		Arc*	first;
		int		index;

		union
		{
			struct
			{
				Node*	tree_parent;
				Node*	tree_child;
				Node*	tree_sibling;
			};
			struct
			{
				GH_REAL w_down;
			};
		};
		union
		{
			struct // used in GomoryHuNested
			{
				bool	is_cut;
				Node*	next; // next node with the same index
			};
			struct // used in GraphWithTreeAndLCA
			{
				Arc*	a_contraction;
				int		depth;
			};
		};
	};
	struct Arc
	{
		Node*	head;
		Arc*	next;
		Arc*	rev;
		GH_REAL	w;
	};

	int		tree_edge_num;
	int		node_num;
	Node*	nodes;
	Node*	root;
	int*	postorder;
	int		i_current;

	Buffer	buf;
	ReusableBuffer	rbuf;

	void FinalizeTree();
};

inline GraphWithTree::GraphWithTree(int _node_num)
	: node_num(_node_num), buf(1024), postorder(NULL), tree_edge_num(0), root(NULL)
{
	nodes = new Node[node_num];
	memset(nodes, 0, node_num*sizeof(Node));
	i_current = -1;
}

inline GraphWithTree::~GraphWithTree()
{
	delete [] nodes;
	if (postorder) delete [] postorder;
}

inline void GraphWithTree::AddEdge(int _i, int _j, GH_REAL w)
{
	Node* i = nodes + _i;
	Node* j = nodes + _j;
	Arc* a = (Arc*) buf.Alloc(2*sizeof(Arc));
	Arc* a_rev = a+1;
	a->rev = a_rev; a_rev->rev = a;
	a->w = w;
	a_rev->w = w;
	a->head = j;
	a_rev->head = i;
	a->next = i->first;
	i->first = a;
	a_rev->next = j->first;
	j->first = a_rev;
}

inline void GraphWithTree::AddTreeEdge(int _i, int _p)
{
	Node* i = nodes + _i;
	Node* p = nodes + _p;
	i->tree_parent = p;
	i->tree_sibling = p->tree_child;
	p->tree_child = i;

	tree_edge_num ++;
	if (root && tree_edge_num == node_num-1) FinalizeTree();
}

void GraphWithTree::AddRoot(int r)
{
	root = nodes + r;

	if (root && tree_edge_num == node_num-1) FinalizeTree();
}

void GraphWithTree::FinalizeTree()
{
	//assert(!postorder);

	if (node_num == 0) return;
	postorder = new int[node_num];
	if (node_num == 1) { postorder[0] = 0; return; }

	Node* i = root;
	int k = 0;
	while ( 1 )
	{
		if (i->tree_child)
		{
			i = i->tree_child;
			continue;
		}
		while ( 1 )
		{
			postorder[k ++] = (int)(i-nodes);
			if (i->tree_sibling) break;
			i = i->tree_parent;
			if (i == root) { postorder[k ++] = (int)(i-nodes); break; }
		}
		if (i == root) break;
		i = i->tree_sibling;
	}
}

inline void GraphWithTree::FinishNode()
{
	Arc* a;
	for (a=nodes[i_current].first; a; a=a->next)
	{
		a->head->a_contraction = NULL;
		a->rev->w = a->w;
	}
}

inline bool GraphWithTree::AddEdge(int j, GH_REAL w)
{
	if (nodes[j].a_contraction)
	{
		nodes[j].a_contraction->w += w;
		return false;
	}
	else
	{
		AddEdge(i_current, j, w);
		nodes[j].a_contraction = nodes[i_current].first;
		return true;
	}
}

inline void GraphWithTree::MergeParallelEdges()
{
	Node* i;
	Arc* a;
	for (i=nodes; i<nodes+node_num; i++)
	{
		Arc** a_prev_ptr = &i->first;
		while ((a=*a_prev_ptr))
		{
			if (a->rev && a->head->a_contraction == NULL)
			{
				a->head->a_contraction = a;
				a_prev_ptr = &a->next;
			}
			else
			{
				if (a->rev)
				{
					a->head->a_contraction->w += a->w;
					a->rev->rev = NULL;
				}
				*a_prev_ptr = a->next;
			}
		}
		for (a=i->first; a; a=a->next)
		{
			a->head->a_contraction->rev->w = a->head->a_contraction->w;
			a->head->a_contraction = NULL;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

class GraphWithTreeAndLCA : public GraphWithTree
{
public:
	GraphWithTreeAndLCA(int n);
	~GraphWithTreeAndLCA() {}
	void SetLCA();
	GraphWithTree* TransformGraph();

	Node* getLCA(Node* i, Node* j)
	{
		if (i->depth < j->depth) { Node* k = i; i = j; j = k; }
		while (i->depth > j->depth) i = i->tree_parent;
		while (i != j) { i = i->tree_parent; j = j->tree_parent; }
		return i;
	}
	Node* getPenultimateNodes(Node*& _i, Node*& _j)
	{
		if (_i == _j) return _i;

		bool swap;
		Node* i;
		Node* j;
		if (_i->depth > _j->depth) { i = _i; j = _j; swap = false; }
		else                       { i = _j; j = _i; swap = true; }
		while (i->depth > j->depth+1) i = i->tree_parent;
		if (i->depth > j->depth)
		{
			if (i->tree_parent == j)
			{
				if (swap) { _i = j; _j = i; }
				else      { _i = i; _j = j; }
				return j;
			}
			i = i->tree_parent;
		}
		while (i->tree_parent != j->tree_parent) { i = i->tree_parent; j = j->tree_parent; }
		if (swap) { _i = j; _j = i; }
		else      { _i = i; _j = j; }
		return i->tree_parent;
	}
	// i must be a predecessor of p
	void AddDelta(Node* i, Node* p, GH_REAL w, GraphWithTree* h)
	{
		while ( i != p )
		{
			h->nodes[(int)(i-nodes)].w_down += w;
			i = i->tree_parent;
		}
	}
};

GraphWithTreeAndLCA::GraphWithTreeAndLCA(int n)
	: GraphWithTree(n)
{
}

void GraphWithTreeAndLCA::SetLCA()
{
	int _i;
	nodes[postorder[node_num-1]].depth = 0;
	for (_i=node_num-2; _i>=0; _i--)
	{
		Node* i = nodes + postorder[_i];
		i->depth = i->tree_parent->depth + 1;
	}
}

GraphWithTree* GraphWithTreeAndLCA::TransformGraph()
{
	Node* i;
	Arc* a;
	GraphWithTree* h = new GraphWithTree(node_num);

	for (i=nodes; i<nodes+node_num; i++)
	{
		for (a=i->first; a; a=a->next)
		{
			Node* j = a->head;
			if (i >= j) continue;
			Node* ii = i;
			Node* jj = j;
			Node* p = getPenultimateNodes(ii, jj);
			if (jj == p)
			{
				AddDelta(i, ii, a->w, h);
			}
			else if (ii == p)
			{
				AddDelta(j, jj, a->w, h);
			}
			else
			{
				h->AddEdge((int)(ii-nodes), (int)(jj-nodes), a->w);
				AddDelta(i, ii, a->w, h);
				AddDelta(j, jj, a->w, h);
			}
		}
	}
	h->MergeParallelEdges();
	return h;
}

#endif

